###############################################################################
#
# Helper functions for testing XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import re
import os.path
from zipfile import ZipFile
from zipfile import BadZipFile
from zipfile import LargeZipFile


def _xml_to_list(xml_str):
    # Convert test generated XML strings into lists for comparison testing.

    # Split the XML string at tag boundaries.
    parser = re.compile(r">\s*<")
    elements = parser.split(xml_str.strip())

    elements = [s.replace("\r", "") for s in elements]

    # Add back the removed brackets.
    for index, element in enumerate(elements):
        if not element[0] == "<":
            elements[index] = "<" + elements[index]
        if not element[-1] == ">":
            elements[index] = elements[index] + ">"

    return elements


def _vml_to_list(vml_str):
    # Convert an Excel generated VML string into a list for comparison testing.
    #
    # The VML data in the testcases is taken from Excel 2007 files. The data
    # has to be massaged significantly to make it suitable for comparison.
    #
    # The VML produced by XlsxWriter can be parsed as ordinary XML.
    vml_str = vml_str.replace("\r", "")

    vml = vml_str.split("\n")
    vml_str = ""

    for line in vml:
        # Skip blank lines.
        if not line:
            continue

        # Strip leading and trailing whitespace.
        line = line.strip()

        # Convert VMLs attribute quotes.
        line = line.replace("'", '"')

        # Add space between attributes.
        if re.search('"$', line):
            line += " "

        # Add newline after element end.
        if re.search(">$", line):
            line += "\n"

        # Split multiple elements.
        line = line.replace("><", ">\n<")

        # Put all of Anchor on one line.
        if line == "<x:Anchor>\n":
            line = line.strip()

        vml_str += line

    # Remove the final newline.
    vml_str = vml_str.rstrip()

    return vml_str.split("\n")


def _sort_rel_file_data(xml_elements):
    # Re-order the relationship elements in an array of XLSX XML rel
    # (relationship) data. This is necessary for comparison since
    # Excel can produce the elements in a semi-random order.

    # We don't want to sort the first or last elements.
    first = xml_elements.pop(0)
    last = xml_elements.pop()

    # Sort the relationship elements.
    xml_elements.sort()

    # Add back the first and last elements.
    xml_elements.insert(0, first)
    xml_elements.append(last)

    return xml_elements


def _compare_xlsx_files(got_file, exp_file, ignore_files, ignore_elements):
    # Compare two XLSX files by extracting the XML files from each
    # zip archive and comparing them.
    #
    # This is used to compare an "expected" file produced by Excel
    # with a "got" file produced by XlsxWriter.
    #
    # In order to compare the XLSX files we convert the data in each
    # XML file into an list of XML elements.
    try:
        # Open the XlsxWriter as a zip file for testing.
        got_zip = ZipFile(got_file, "r")
    except IOError as e:
        error = "XlsxWriter file error: " + str(e)
        return error, ""
    except (BadZipFile, LargeZipFile) as e:
        error = "XlsxWriter zipfile error, '" + got_file + "': " + str(e)
        return error, ""

    try:
        # Open the Excel as a zip file for testing.
        exp_zip = ZipFile(exp_file, "r")
    except IOError as e:
        error = "Excel file error: " + str(e)
        return error, ""
    except (BadZipFile, LargeZipFile) as e:
        error = "Excel zipfile error, '" + exp_file + "': " + str(e)
        return error, ""

    # Get the filenames from the zip files.
    got_files = sorted(got_zip.namelist())
    exp_files = sorted(exp_zip.namelist())

    # Ignore some test specific filenames.
    got_files = [name for name in got_files if name not in ignore_files]
    exp_files = [name for name in exp_files if name not in ignore_files]

    # Check that each XLSX container has the same files.
    if got_files != exp_files:
        return got_files, exp_files

    # Compare each file in the XLSX containers.
    for filename in exp_files:
        got_xml_str = got_zip.read(filename)
        exp_xml_str = exp_zip.read(filename)

        # Compare binary files with string comparison based on extension.
        extension = os.path.splitext(filename)[1]
        if extension in (".png", ".jpeg", ".gif", ".bmp", ".wmf", ".emf", ".bin"):
            if got_xml_str != exp_xml_str:
                return "got: %s" % filename, "exp: %s" % filename
            continue

        got_xml_str = got_xml_str.decode("utf-8")
        exp_xml_str = exp_xml_str.decode("utf-8")

        # Remove dates and user specific data from the core.xml data.
        if filename == "docProps/core.xml":
            exp_xml_str = re.sub(r" ?John", "", exp_xml_str)
            exp_xml_str = re.sub(
                r"\d\d\d\d-\d\d-\d\dT\d\d\:\d\d:\d\dZ", "", exp_xml_str
            )
            got_xml_str = re.sub(
                r"\d\d\d\d-\d\d-\d\dT\d\d\:\d\d:\d\dZ", "", got_xml_str
            )

        # Remove workbookView dimensions which are almost always different
        # and calcPr which can have different Excel version ids.
        if filename == "xl/workbook.xml":
            exp_xml_str = re.sub(r"<workbookView[^>]*>", "<workbookView/>", exp_xml_str)
            got_xml_str = re.sub(r"<workbookView[^>]*>", "<workbookView/>", got_xml_str)
            exp_xml_str = re.sub(r"<calcPr[^>]*>", "<calcPr/>", exp_xml_str)
            got_xml_str = re.sub(r"<calcPr[^>]*>", "<calcPr/>", got_xml_str)

        # Remove printer specific settings from Worksheet pageSetup elements.
        if re.match(r"xl/worksheets/sheet\d.xml", filename):
            exp_xml_str = re.sub(r'horizontalDpi="200" ', "", exp_xml_str)
            exp_xml_str = re.sub(r'verticalDpi="200" ', "", exp_xml_str)
            exp_xml_str = re.sub(r'(<pageSetup[^>]*) r:id="rId1"', r"\1", exp_xml_str)

        # Remove Chart pageMargin dimensions which are almost always different.
        if re.match(r"xl/charts/chart\d.xml", filename):
            exp_xml_str = re.sub(
                r"<c:pageMargins[^>]*>", "<c:pageMargins/>", exp_xml_str
            )
            got_xml_str = re.sub(
                r"<c:pageMargins[^>]*>", "<c:pageMargins/>", got_xml_str
            )

        # Convert the XML string to lists for comparison.
        if re.search(".vml$", filename):
            got_xml = _xml_to_list(got_xml_str)
            exp_xml = _vml_to_list(exp_xml_str)
        else:
            got_xml = _xml_to_list(got_xml_str)
            exp_xml = _xml_to_list(exp_xml_str)

        # Ignore test specific XML elements for defined filenames.
        if filename in ignore_elements:
            patterns = ignore_elements[filename]

            for pattern in patterns:
                exp_xml = [tag for tag in exp_xml if not re.match(pattern, tag)]
                got_xml = [tag for tag in got_xml if not re.match(pattern, tag)]

        # Reorder the XML elements in the XLSX relationship files.
        if filename == "[Content_Types].xml" or re.search(".rels$", filename):
            got_xml = _sort_rel_file_data(got_xml)
            exp_xml = _sort_rel_file_data(exp_xml)

        # Compared the XML elements in each file.
        if got_xml != exp_xml:
            got_xml.insert(0, filename)
            exp_xml.insert(0, filename)
            return got_xml, exp_xml

    # If we got here the files are the same.
    return "Ok", "Ok"


# External wrapper function to allow simplified equality testing of two Excel
# files. Note, this function doesn't test equivalence, only equality.
def compare_xlsx_files(file1, file2, ignore_files=None, ignore_elements=None):
    if ignore_files is None:
        ignore_files = []

    if ignore_elements is None:
        ignore_elements = []

    got, exp = _compare_xlsx_files(file1, file2, ignore_files, ignore_elements)

    return got == exp



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ..helperfunctions import _xml_to_list
from ...worksheet import Worksheet


class TestAssembleWorksheet(unittest.TestCase):
    """
    Test assembling a complete Worksheet file.

    """

    def test_assemble_xml_file(self):
        """Test writing a worksheet with conditional formatting."""
        self.maxDiff = None

        fh = StringIO()
        worksheet = Worksheet()
        worksheet._set_filehandle(fh)
        worksheet.select()

        worksheet.write("A1", 1)
        worksheet.write("A2", 2)
        worksheet.write("A3", 3)
        worksheet.write("A4", 4)
        worksheet.write("A5", 5)
        worksheet.write("A6", 6)
        worksheet.write("A7", 7)
        worksheet.write("A8", 8)
        worksheet.write("A9", 9)
        worksheet.write("A10", 10)
        worksheet.write("A11", 11)
        worksheet.write("A12", 12)

        worksheet.conditional_format(
            "A1:A12",
            {
                "type": "data_bar",
                "min_value": 5,
                "mid_value": 52,  # Should be ignored.
                "max_value": 90,
                "min_length": 5,
                "max_length": 95,
                "min_type": "num",
                "mid_type": "percentile",  # Should be ignored.
                "max_type": "percent",
                "bar_color": "#8DB4E3",
            },
        )

        worksheet._assemble_xml_file()

        exp = _xml_to_list(
            """
                <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <worksheet xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">
                  <dimension ref="A1:A12"/>
                  <sheetViews>
                    <sheetView tabSelected="1" workbookViewId="0"/>
                  </sheetViews>
                  <sheetFormatPr defaultRowHeight="15"/>
                  <sheetData>
                    <row r="1" spans="1:1">
                      <c r="A1">
                        <v>1</v>
                      </c>
                    </row>
                    <row r="2" spans="1:1">
                      <c r="A2">
                        <v>2</v>
                      </c>
                    </row>
                    <row r="3" spans="1:1">
                      <c r="A3">
                        <v>3</v>
                      </c>
                    </row>
                    <row r="4" spans="1:1">
                      <c r="A4">
                        <v>4</v>
                      </c>
                    </row>
                    <row r="5" spans="1:1">
                      <c r="A5">
                        <v>5</v>
                      </c>
                    </row>
                    <row r="6" spans="1:1">
                      <c r="A6">
                        <v>6</v>
                      </c>
                    </row>
                    <row r="7" spans="1:1">
                      <c r="A7">
                        <v>7</v>
                      </c>
                    </row>
                    <row r="8" spans="1:1">
                      <c r="A8">
                        <v>8</v>
                      </c>
                    </row>
                    <row r="9" spans="1:1">
                      <c r="A9">
                        <v>9</v>
                      </c>
                    </row>
                    <row r="10" spans="1:1">
                      <c r="A10">
                        <v>10</v>
                      </c>
                    </row>
                    <row r="11" spans="1:1">
                      <c r="A11">
                        <v>11</v>
                      </c>
                    </row>
                    <row r="12" spans="1:1">
                      <c r="A12">
                        <v>12</v>
                      </c>
                    </row>
                  </sheetData>
                  <conditionalFormatting sqref="A1:A12">
                    <cfRule type="dataBar" priority="1">
                      <dataBar minLength="5" maxLength="95">
                        <cfvo type="num" val="5"/>
                        <cfvo type="percent" val="90"/>
                        <color rgb="FF8DB4E3"/>
                      </dataBar>
                    </cfRule>
                  </conditionalFormatting>
                  <pageMargins left="0.7" right="0.7" top="0.75" bottom="0.75" header="0.3" footer="0.3"/>
                </worksheet>
                """
        )

        got = _xml_to_list(fh.getvalue())

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("format12.xlsx")

    def test_create_file(self):
        """Test a vertical and horizontal centered format."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        top_left_bottom = workbook.add_format(
            {
                "left": 1,
                "top": 1,
                "bottom": 1,
            }
        )

        top_bottom = workbook.add_format(
            {
                "top": 1,
                "bottom": 1,
            }
        )

        top_left = workbook.add_format(
            {
                "left": 1,
                "top": 1,
            }
        )

        worksheet.write("B2", "test", top_left_bottom)
        worksheet.write("D2", "test", top_left)
        worksheet.write("F2", "test", top_bottom)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("properties01.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        workbook.set_properties(
            {
                "title": "This is an example spreadsheet",
                "subject": "With document properties",
                "author": "Someone",
                "manager": "Dr. Heinz Doofenshmirtz",
                "company": "of Wolves",
                "category": "Example spreadsheets",
                "keywords": "Sample, Example, Properties",
                "comments": "Created with Perl and Excel::Writer::XLSX",
                "status": "Quo",
            }
        )

        worksheet.set_column("A:A", 70)
        worksheet.write(
            "A1",
            "Select 'Office Button -> Prepare -> Properties' to see the file properties.",
        )

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWriteAutoFilter(unittest.TestCase):
    """
    Test the Worksheet _write_auto_filter() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)
        self.worksheet.name = "Sheet1"
        self.worksheet.autofilter("A1:D51")

    def test_write_auto_filter_1(self):
        """Test the _write_auto_filter() method"""

        self.worksheet._write_auto_filter()

        exp = """<autoFilter ref="A1:D51"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_2(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x == East"
        exp = """<autoFilter ref="A1:D51"><filterColumn colId="0"><filters><filter val="East"/></filters></filterColumn></autoFilter>"""

        self.worksheet.filter_column(0, filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_3(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x == East or  x == North"
        exp = """<autoFilter ref="A1:D51"><filterColumn colId="0"><filters><filter val="East"/><filter val="North"/></filters></filterColumn></autoFilter>"""

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_4(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x == East and x == North"
        exp = """<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters and="1"><customFilter val="East"/><customFilter val="North"/></customFilters></filterColumn></autoFilter>"""

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_5(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x != East"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters><customFilter operator="notEqual" val="East"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_6(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x == S*"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters><customFilter val="S*"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_7(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x != S*"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters><customFilter operator="notEqual" val="S*"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_8(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x == *h"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters><customFilter val="*h"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_9(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x != *h"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters><customFilter operator="notEqual" val="*h"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_10(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x =~ *o*"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters><customFilter val="*o*"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_11(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x !~ *r*"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><customFilters><customFilter operator="notEqual" val="*r*"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("A", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_12(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x == 1000"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="2"><filters><filter val="1000"/></filters></filterColumn></autoFilter>'

        self.worksheet.filter_column(2, filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_13(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x != 2000"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="2"><customFilters><customFilter operator="notEqual" val="2000"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("C", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_14(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x > 3000"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="2"><customFilters><customFilter operator="greaterThan" val="3000"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("C", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_15(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x >= 4000"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="2"><customFilters><customFilter operator="greaterThanOrEqual" val="4000"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("C", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_16(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x < 5000"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="2"><customFilters><customFilter operator="lessThan" val="5000"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("C", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_17(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x <= 6000"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="2"><customFilters><customFilter operator="lessThanOrEqual" val="6000"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("C", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_18(self):
        """Test the _write_auto_filter() method"""

        filter_condition = "x >= 1000 and x <= 2000"
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="2"><customFilters and="1"><customFilter operator="greaterThanOrEqual" val="1000"/><customFilter operator="lessThanOrEqual" val="2000"/></customFilters></filterColumn></autoFilter>'

        self.worksheet.filter_column("C", filter_condition)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_19(self):
        """Test the _write_auto_filter() method"""

        matches = ["East"]
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><filters><filter val="East"/></filters></filterColumn></autoFilter>'

        self.worksheet.filter_column_list("A", matches)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_20(self):
        """Test the _write_auto_filter() method"""

        matches = ["East", "North"]
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="0"><filters><filter val="East"/><filter val="North"/></filters></filterColumn></autoFilter>'

        self.worksheet.filter_column_list("A", matches)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_auto_filter_21(self):
        """Test the _write_auto_filter() method"""

        matches = ["February", "January", "July", "June"]
        exp = '<autoFilter ref="A1:D51"><filterColumn colId="3"><filters><filter val="February"/><filter val="January"/><filter val="July"/><filter val="June"/></filters></filterColumn></autoFilter>'

        self.worksheet.filter_column_list(3, matches)
        self.worksheet._write_auto_filter()

        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...vml import Vml


class TestWriteVshapetype(unittest.TestCase):
    """
    Test the Vml _write_shapetype() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.vml = Vml()
        self.vml._set_filehandle(self.fh)

    def test_write_comment_shapetype(self):
        """Test the _write_comment_shapetype() method"""

        self.vml._write_comment_shapetype()

        exp = """<v:shapetype id="_x0000_t202" coordsize="21600,21600" o:spt="202" path="m,l,21600r21600,l21600,xe"><v:stroke joinstyle="miter"/><v:path gradientshapeok="t" o:connecttype="rect"/></v:shapetype>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_button_shapetype(self):
        """Test the _write_button_shapetype() method"""

        self.vml._write_button_shapetype()

        exp = """<v:shapetype id="_x0000_t201" coordsize="21600,21600" o:spt="201" path="m,l,21600r21600,l21600,xe"><v:stroke joinstyle="miter"/><v:path shadowok="f" o:extrusionok="f" strokeok="f" fillok="f" o:connecttype="rect"/><o:lock v:ext="edit" shapetype="t"/></v:shapetype>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("optimize10.xlsx")
        self.set_text_file("unicode_polish_utf8.txt")

    def test_create_file(self):
        """Test example file converting Unicode text."""

        # Open the input file with the correct encoding.
        textfile = open(self.txt_filename, mode="r", encoding="utf-8")

        # Create an new Excel file and convert the text data.
        workbook = Workbook(
            self.got_filename, {"constant_memory": True, "in_memory": False}
        )
        worksheet = workbook.add_worksheet()

        # Widen the first column to make the text clearer.
        worksheet.set_column("A:A", 50)

        # Start from the first cell.
        row = 0
        col = 0

        # Read the text file and write it to the worksheet.
        for line in textfile:
            # Ignore the comments in the sample file.
            if line.startswith("#"):
                continue

            # Write any other lines to the worksheet.
            worksheet.write(row, col, line.rstrip("\n"))
            row += 1

        workbook.close()
        textfile.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...drawing import Drawing


class TestWriteXdrcNvPr(unittest.TestCase):
    """
    Test the Drawing _write_c_nv_pr() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.drawing = Drawing()
        self.drawing._set_filehandle(self.fh)

    def test_write_c_nv_pr(self):
        """Test the _write_c_nv_pr() method"""

        self.drawing._write_c_nv_pr(2, "Chart 1", None, None, None, None)

        exp = """<xdr:cNvPr id="2" name="Chart 1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

        self.drawing._write_c_nv_pr(2, "Chart 1", None, 1, "tip", None)

        exp = """<xdr:cNvPr id="2" name="Chart 1"/><xdr:cNvPr id="2" name="Chart 1"><a:hlinkClick xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" r:id="rId1" tooltip="tip"/></xdr:cNvPr>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("chart_bar22.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()
        chart = workbook.add_chart({"type": "bar"})

        chart.axis_ids = [43706240, 43727104]

        headers = ["Series 1", "Series 2", "Series 3"]

        data = [
            ["Category 1", "Category 2", "Category 3", "Category 4"],
            [4.3, 2.5, 3.5, 4.5],
            [2.4, 4.5, 1.8, 2.8],
            [2, 2, 3, 5],
        ]

        worksheet.set_column("A:D", 11)

        worksheet.write_row("B1", headers)
        worksheet.write_column("A2", data[0])
        worksheet.write_column("B2", data[1])
        worksheet.write_column("C2", data[2])
        worksheet.write_column("D2", data[3])

        chart.add_series(
            {
                "categories": "=Sheet1!$A$2:$A$5",
                "values": "=Sheet1!$B$2:$B$5",
                "categories_data": data[0],
                "values_data": data[1],
            }
        )

        chart.add_series(
            {
                "categories": "=Sheet1!$A$2:$A$5",
                "values": "=Sheet1!$C$2:$C$5",
                "categories_data": data[0],
                "values_data": data[2],
            }
        )

        chart.add_series(
            {
                "categories": "=Sheet1!$A$2:$A$5",
                "values": "=Sheet1!$D$2:$D$5",
                "categories_data": data[0],
                "values_data": data[3],
            }
        )

        worksheet.insert_chart("E9", chart)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWriteRowBreaks(unittest.TestCase):
    """
    Test the Worksheet _write_row_breaks() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_row_breaks_1(self):
        """Test the _write_row_breaks() method"""

        self.worksheet.hbreaks = [1]

        self.worksheet._write_row_breaks()

        exp = """<rowBreaks count="1" manualBreakCount="1"><brk id="1" max="16383" man="1"/></rowBreaks>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_row_breaks_2(self):
        """Test the _write_row_breaks() method"""

        self.worksheet.hbreaks = [15, 7, 3, 0]

        self.worksheet._write_row_breaks()

        exp = """<rowBreaks count="3" manualBreakCount="3"><brk id="3" max="16383" man="1"/><brk id="7" max="16383" man="1"/><brk id="15" max="16383" man="1"/></rowBreaks>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("formula_results01.xlsx")

        self.ignore_files = [
            "xl/calcChain.xml",
            "[Content_Types].xml",
            "xl/_rels/workbook.xml.rels",
        ]

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with formula errors."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.write_formula("A1", "1+1", None, 2)
        worksheet.write_formula("A2", '"Foo"', None, "Foo")
        worksheet.write_formula("A3", "IF(B3,FALSE,TRUE)", None, True)
        worksheet.write_formula("A4", "IF(B4,TRUE,FALSE)", None, False)
        worksheet.write_formula("A5", "#DIV/0!", None, "#DIV/0!")
        worksheet.write_formula("A6", "#N/A", None, "#N/A")
        worksheet.write_formula("A7", "#NAME?", None, "#NAME?")
        worksheet.write_formula("A8", "#NULL!", None, "#NULL!")
        worksheet.write_formula("A9", "#NUM!", None, "#NUM!")
        worksheet.write_formula("A10", "#REF!", None, "#REF!")
        worksheet.write_formula("A11", "#VALUE!", None, "#VALUE!")
        worksheet.write_formula("A12", "1/0", None, "#DIV/0!")

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...chartsheet import Chartsheet


class TestInitialisation(unittest.TestCase):
    """
    Test initialisation of the Chartsheet class and call a method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.chartsheet = Chartsheet()
        self.chartsheet._set_filehandle(self.fh)

    def test_xml_declaration(self):
        """Test Chartsheet xml_declaration()"""

        self.chartsheet._xml_declaration()

        exp = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...styles import Styles


class TestWriteCellStyles(unittest.TestCase):
    """
    Test the Styles _write_cell_styles() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.styles = Styles()
        self.styles._set_filehandle(self.fh)

    def test_write_cell_styles(self):
        """Test the _write_cell_styles() method"""

        self.styles._write_cell_styles()

        exp = """<cellStyles count="1"><cellStyle name="Normal" xfId="0" builtinId="0"/></cellStyles>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...styles import Styles


class TestWriteColors(unittest.TestCase):
    """
    Test the Styles _write_colors() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.styles = Styles()
        self.styles._set_filehandle(self.fh)

    def test_write_colors1(self):
        """Test the _write_colors() method"""

        self.styles.custom_colors = ["FF26DA55"]
        self.styles._write_colors()

        exp = """<colors><mruColors><color rgb="FF26DA55"/></mruColors></colors>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_colors2(self):
        """Test the _write_colors() method"""

        self.styles.custom_colors = ["FF26DA55", "FF792DC8", "FF646462"]
        self.styles._write_colors()

        exp = """<colors><mruColors><color rgb="FF646462"/><color rgb="FF792DC8"/><color rgb="FF26DA55"/></mruColors></colors>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_colors3(self):
        """Test the _write_colors() method"""

        self.styles.custom_colors = [
            "FF792DC8",
            "FF646462",
            "FF5EA29C",
            "FF583AC6",
            "FFE31DAF",
            "FFA1A759",
            "FF600FF1",
            "FF0CF49C",
            "FFE3FA06",
            "FF913AC6",
            "FFB97847",
            "FFD97827",
        ]

        self.styles._write_colors()

        exp = """<colors><mruColors><color rgb="FFD97827"/><color rgb="FFB97847"/><color rgb="FF913AC6"/><color rgb="FFE3FA06"/><color rgb="FF0CF49C"/><color rgb="FF600FF1"/><color rgb="FFA1A759"/><color rgb="FFE31DAF"/><color rgb="FF583AC6"/><color rgb="FF5EA29C"/></mruColors></colors>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ..helperfunctions import _xml_to_list
from ...table import Table
from ...worksheet import Worksheet
from ...workbook import WorksheetMeta
from ...sharedstrings import SharedStringTable


class TestAssembleTable(unittest.TestCase):
    """
    Test assembling a complete Table file.

    """

    def test_assemble_xml_file(self):
        """Test writing a table"""
        self.maxDiff = None

        worksheet = Worksheet()
        worksheet.worksheet_meta = WorksheetMeta()
        worksheet.str_table = SharedStringTable()

        worksheet.add_table("C4:F13", {"header_row": False})
        worksheet._prepare_tables(1, {})

        fh = StringIO()
        table = Table()
        table._set_filehandle(fh)

        table._set_properties(worksheet.tables[0])

        table._assemble_xml_file()

        exp = _xml_to_list(
            """
                <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <table xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" id="1" name="Table1" displayName="Table1" ref="C4:F13" headerRowCount="0" totalsRowShown="0">
                  <tableColumns count="4">
                    <tableColumn id="1" name="Column1"/>
                    <tableColumn id="2" name="Column2"/>
                    <tableColumn id="3" name="Column3"/>
                    <tableColumn id="4" name="Column4"/>
                  </tableColumns>
                  <tableStyleInfo name="TableStyleMedium9" showFirstColumn="0" showLastColumn="0" showRowStripes="1" showColumnStripes="0"/>
                </table>
                """
        )

        got = _xml_to_list(fh.getvalue())

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
import warnings
from ...utility import xl_rowcol_to_cell
from ...utility import xl_rowcol_to_cell_fast


class TestUtility(unittest.TestCase):
    """
    Test xl_rowcol_to_cell() utility function.

    """

    def test_xl_rowcol_to_cell(self):
        """Test xl_rowcol_to_cell()"""

        tests = [
            # row, col, A1 string
            (0, 0, "A1"),
            (0, 1, "B1"),
            (0, 2, "C1"),
            (0, 9, "J1"),
            (1, 0, "A2"),
            (2, 0, "A3"),
            (9, 0, "A10"),
            (1, 24, "Y2"),
            (7, 25, "Z8"),
            (9, 26, "AA10"),
            (1, 254, "IU2"),
            (1, 255, "IV2"),
            (1, 256, "IW2"),
            (0, 16383, "XFD1"),
            (1048576, 16384, "XFE1048577"),
            (0, -1, None),
            (-1, 0, None),
        ]

        for row, col, range in tests:
            exp = range
            got = xl_rowcol_to_cell(row, col)

            # Ignore the warnings for negative values.
            warnings.filterwarnings("ignore")

            self.assertEqual(got, exp)

    def test_xl_rowcol_to_cell_abs(self):
        """Test xl_rowcol_to_cell() with absolute references"""

        tests = [
            # row, col, row_abs, col_abs, A1 range
            (0, 0, True, False, "A$1"),
            (0, 0, False, True, "$A1"),
            (0, 0, True, True, "$A$1"),
            (-1, 0, 0, 0, None),
            (0, -1, 0, 0, None),
        ]

        for row, col, row_abs, col_abs, range in tests:
            exp = range
            got = xl_rowcol_to_cell(row, col, row_abs, col_abs)

            # Ignore the warnings for negative values.
            warnings.filterwarnings("ignore")

            self.assertEqual(got, exp)

    def test_xl_rowcol_to_cell_fast(self):
        """Test xl_rowcol_to_cell_fast()"""

        tests = [
            # row, col, A1 range
            (0, 0, "A1"),
            (0, 1, "B1"),
            (0, 2, "C1"),
            (0, 9, "J1"),
            (1, 0, "A2"),
            (2, 0, "A3"),
            (9, 0, "A10"),
            (1, 24, "Y2"),
            (7, 25, "Z8"),
            (9, 26, "AA10"),
            (1, 254, "IU2"),
            (1, 255, "IV2"),
            (1, 256, "IW2"),
            (0, 16383, "XFD1"),
            (1048576, 16384, "XFE1048577"),
        ]

        for row, col, range in tests:
            exp = range
            got = xl_rowcol_to_cell_fast(row, col)
            self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWriteSheetViews(unittest.TestCase):
    """
    Test the Worksheet _write_sheet_views() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_sheet_views(self):
        """Test the _write_sheet_views() method"""

        self.worksheet.select()
        self.worksheet._write_sheet_views()

        exp = """<sheetViews><sheetView tabSelected="1" workbookViewId="0"/></sheetViews>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views_zoom_100(self):
        """Test the _write_sheet_views() method"""

        self.worksheet.select()
        self.worksheet.set_zoom(100)  # Default. Should be ignored.
        self.worksheet._write_sheet_views()

        exp = """<sheetViews><sheetView tabSelected="1" workbookViewId="0"/></sheetViews>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views_zoom_200(self):
        """Test the _write_sheet_views() method"""

        self.worksheet.select()
        self.worksheet.set_zoom(200)
        self.worksheet._write_sheet_views()

        exp = """<sheetViews><sheetView tabSelected="1" zoomScale="200" zoomScaleNormal="200" workbookViewId="0"/></sheetViews>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views_right_to_left(self):
        """Test the _write_sheet_views() method"""

        self.worksheet.select()
        self.worksheet.right_to_left()
        self.worksheet._write_sheet_views()

        exp = """<sheetViews><sheetView rightToLeft="1" tabSelected="1" workbookViewId="0"/></sheetViews>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views_hide_zero(self):
        """Test the _write_sheet_views() method"""

        self.worksheet.select()
        self.worksheet.hide_zero()
        self.worksheet._write_sheet_views()

        exp = """<sheetViews><sheetView showZeros="0" tabSelected="1" workbookViewId="0"/></sheetViews>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views_page_view(self):
        """Test the _write_sheet_views() method"""

        self.worksheet.select()
        self.worksheet.set_page_view()
        self.worksheet._write_sheet_views()

        exp = """<sheetViews><sheetView tabSelected="1" view="pageLayout" workbookViewId="0"/></sheetViews>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWriteSheetViews(unittest.TestCase):
    """
    Test the Worksheet _write_sheet_views() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_sheet_views1(self):
        """Test the _write_sheet_views() method with freeze panes"""

        self.worksheet.select()

        self.worksheet.set_selection("A2")
        self.worksheet.freeze_panes(1, 0, 20, 0)

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><pane ySplit="1" topLeftCell="A21" activePane="bottomLeft" state="frozen"/><selection pane="bottomLeft" activeCell="A2" sqref="A2"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views2(self):
        """Test the _write_sheet_views() method with freeze panes"""

        self.worksheet.select()

        self.worksheet.set_selection("A1")
        self.worksheet.freeze_panes(1, 0, 20, 0)

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><pane ySplit="1" topLeftCell="A21" activePane="bottomLeft" state="frozen"/><selection pane="bottomLeft"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views3(self):
        """Test the _write_sheet_views() method with freeze panes"""

        self.worksheet.select()

        self.worksheet.set_selection("B1")
        self.worksheet.freeze_panes(0, 1, 0, 4)

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><pane xSplit="1" topLeftCell="E1" activePane="topRight" state="frozen"/><selection pane="topRight" activeCell="B1" sqref="B1"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views4(self):
        """Test the _write_sheet_views() method with freeze panes"""

        self.worksheet.select()

        self.worksheet.freeze_panes(0, 1, 0, 4)

        self.worksheet.set_selection("A1")
        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><pane xSplit="1" topLeftCell="E1" activePane="topRight" state="frozen"/><selection pane="topRight"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views5(self):
        """Test the _write_sheet_views() method with freeze panes"""

        self.worksheet.select()

        self.worksheet.set_selection("G4")
        self.worksheet.freeze_panes(3, 6, 6, 8)

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><pane xSplit="6" ySplit="3" topLeftCell="I7" activePane="bottomRight" state="frozen"/><selection pane="topRight" activeCell="G1" sqref="G1"/><selection pane="bottomLeft" activeCell="A4" sqref="A4"/><selection pane="bottomRight" activeCell="G4" sqref="G4"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views6(self):
        """Test the _write_sheet_views() method with freeze panes"""

        self.worksheet.select()

        self.worksheet.set_selection("A1")
        self.worksheet.freeze_panes(3, 6, 6, 8)

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><pane xSplit="6" ySplit="3" topLeftCell="I7" activePane="bottomRight" state="frozen"/><selection pane="topRight" activeCell="G1" sqref="G1"/><selection pane="bottomLeft" activeCell="A4" sqref="A4"/><selection pane="bottomRight"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



##############################################################################
#
# Simple Python program to test the speed and memory usage of
# the XlsxWriter module.
#
# python perf_pyx.py [num_rows] [optimization_mode]
#
# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org

import sys
import xlsxwriter
from time import perf_counter
from pympler.asizeof import asizeof

# Default to 1000 rows and non-optimised.
if len(sys.argv) > 1:
    row_max = int(sys.argv[1]) // 2
else:
    row_max = 1000

if len(sys.argv) > 2 and int(sys.argv[2]) == 1:
    optimise = 1
else:
    optimise = 0

if len(sys.argv) > 3 and int(sys.argv[3]) == 1:
    get_memory_size = 1
else:
    get_memory_size = 0


col_max = 50

# Start timing after everything is loaded.
start_time = perf_counter()

# Start of program being tested.
workbook = xlsxwriter.Workbook('py_ewx.xlsx',
                               {'constant_memory': optimise})
worksheet = workbook.add_worksheet()

worksheet.set_column(0, col_max, 18)

for row in range(0, row_max):
    for col in range(0, col_max):
        worksheet.write_string(row * 2, col, "Row: %d Col: %d" % (row, col))
    for col in range(0, col_max + 1):
        worksheet.write_number(row * 2 + 1, col, row + col)

# Get total memory size for workbook object before closing it.
if get_memory_size:
    total_size = asizeof(workbook)
else:
    total_size = 0

workbook.close()

# Get the elapsed time.
elapsed = perf_counter() - start_time

# Print a simple CSV output for reporting.

print("%6d, %3d, %6.2f, %d" % (row_max * 2, col_max, elapsed, total_size))



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...table import Table


class TestWriteTableStyleInfo(unittest.TestCase):
    """
    Test the Table _write_table_style_info() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.table = Table()
        self.table._set_filehandle(self.fh)

    def test_write_table_style_info(self):
        """Test the _write_table_style_info() method"""

        self.table.properties = {
            "style": "TableStyleMedium9",
            "show_first_col": 0,
            "show_last_col": 0,
            "show_row_stripes": 1,
            "show_col_stripes": 0,
        }

        self.table._write_table_style_info()

        exp = """<tableStyleInfo name="TableStyleMedium9" showFirstColumn="0" showLastColumn="0" showRowStripes="1" showColumnStripes="0"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("selection02.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet1 = workbook.add_worksheet()
        worksheet2 = workbook.add_worksheet()
        worksheet3 = workbook.add_worksheet()
        worksheet4 = workbook.add_worksheet()
        worksheet5 = workbook.add_worksheet()
        worksheet6 = workbook.add_worksheet()

        worksheet1.set_selection(3, 2, 3, 2)
        worksheet2.set_selection(3, 2, 6, 6)
        worksheet3.set_selection(6, 6, 3, 2)
        worksheet4.set_selection("C4")
        worksheet5.set_selection("C4:G7")
        worksheet6.set_selection("G7:C4")

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("chart_combined10.xlsx")
        self.ignore_elements = {"xl/charts/chart1.xml": ["<c:dispBlanksAs"]}

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()
        chart_doughnut = workbook.add_chart({"type": "doughnut"})
        chart_pie = workbook.add_chart({"type": "pie"})

        worksheet.write_column("H2", ["Donut", 25, 50, 25, 100])
        worksheet.write_column("I2", ["Pie", 75, 1, 124])

        chart_doughnut.add_series(
            {
                "name": "=Sheet1!$H$2",
                "values": "=Sheet1!$H$3:$H$6",
            }
        )

        chart_pie.add_series(
            {
                "name": "=Sheet1!$I$2",
                "values": "=Sheet1!$I$3:$I$6",
            }
        )

        chart_doughnut.combine(chart_pie)

        worksheet.insert_chart("E9", chart_doughnut)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#
import warnings
from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("simple01.xlsx")

    def test_close_file_twice(self):
        """Test warning when closing workbook more than once."""

        workbook = Workbook(self.got_filename)
        worksheet = workbook.add_worksheet()

        worksheet.write_string(0, 0, "Hello")
        worksheet.write_number(1, 0, 123)

        workbook.close()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            workbook.close()
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("repeat04.xlsx")

        self.ignore_files = [
            "xl/printerSettings/printerSettings1.bin",
            "xl/worksheets/_rels/sheet1.xml.rels",
        ]
        self.ignore_elements = {
            "[Content_Types].xml": ['<Default Extension="bin"'],
            "xl/worksheets/sheet1.xml": ["<pageMargins", "<pageSetup"],
        }

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with repeat rows when the sheet name contains a space."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet("Sheet 1")

        worksheet.repeat_rows(0)

        worksheet.write("A1", "Foo")

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ..helperfunctions import _xml_to_list
from ...worksheet import Worksheet


class TestAssembleWorksheet(unittest.TestCase):
    """
    Test assembling a complete Worksheet file.

    """

    def test_assemble_xml_file(self):
        """Test writing a worksheet with no cell data."""
        self.maxDiff = None

        fh = StringIO()
        worksheet = Worksheet()
        worksheet._set_filehandle(fh)
        worksheet.select()
        worksheet.name = "Sheet1"
        worksheet.excel_version = 2010

        data = [-2, 2, 3, -1, 0]
        worksheet.write_row("A1", data)

        # Set up sparklines.
        worksheet.add_sparkline("F1", {"range": "Sheet1!A1:E1", "max": 4, "min": 0})

        worksheet._assemble_xml_file()

        exp = _xml_to_list(
            """
                <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <worksheet xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:x14ac="http://schemas.microsoft.com/office/spreadsheetml/2009/9/ac" mc:Ignorable="x14ac">
                  <dimension ref="A1:E1"/>
                  <sheetViews>
                    <sheetView tabSelected="1" workbookViewId="0"/>
                  </sheetViews>
                  <sheetFormatPr defaultRowHeight="15" x14ac:dyDescent="0.25"/>
                  <sheetData>
                    <row r="1" spans="1:5" x14ac:dyDescent="0.25">
                      <c r="A1">
                        <v>-2</v>
                      </c>
                      <c r="B1">
                        <v>2</v>
                      </c>
                      <c r="C1">
                        <v>3</v>
                      </c>
                      <c r="D1">
                        <v>-1</v>
                      </c>
                      <c r="E1">
                        <v>0</v>
                      </c>
                    </row>
                  </sheetData>
                  <pageMargins left="0.7" right="0.7" top="0.75" bottom="0.75" header="0.3" footer="0.3"/>
                  <extLst>
                    <ext xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main" uri="{05C60535-1F16-4fd2-B633-F4F36F0B64E0}">
                      <x14:sparklineGroups xmlns:xm="http://schemas.microsoft.com/office/excel/2006/main">
                        <x14:sparklineGroup manualMax="4" manualMin="0" displayEmptyCellsAs="gap" minAxisType="custom" maxAxisType="custom">
                          <x14:colorSeries theme="4" tint="-0.499984740745262"/>
                          <x14:colorNegative theme="5"/>
                          <x14:colorAxis rgb="FF000000"/>
                          <x14:colorMarkers theme="4" tint="-0.499984740745262"/>
                          <x14:colorFirst theme="4" tint="0.39997558519241921"/>
                          <x14:colorLast theme="4" tint="0.39997558519241921"/>
                          <x14:colorHigh theme="4"/>
                          <x14:colorLow theme="4"/>
                          <x14:sparklines>
                            <x14:sparkline>
                              <xm:f>Sheet1!A1:E1</xm:f>
                              <xm:sqref>F1</xm:sqref>
                            </x14:sparkline>
                          </x14:sparklines>
                        </x14:sparklineGroup>
                      </x14:sparklineGroups>
                    </ext>
                  </extLst>
                </worksheet>
                """
        )

        got = _xml_to_list(fh.getvalue())

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from collections import namedtuple
from io import StringIO
from ...worksheet import Worksheet


class TestWriteCell(unittest.TestCase):
    """
    Test the Worksheet _write_cell() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_cell_number(self):
        """Test the _write_cell() method for numbers."""

        cell_tuple = namedtuple("Number", "number, format")
        cell = cell_tuple(1, None)

        self.worksheet._write_cell(0, 0, cell)

        exp = """<c r="A1"><v>1</v></c>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_cell_string(self):
        """Test the _write_cell() method for strings."""

        cell_tuple = namedtuple("String", "string, format")
        cell = cell_tuple(0, None)

        self.worksheet._write_cell(3, 1, cell)

        exp = """<c r="B4" t="s"><v>0</v></c>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_cell_formula01(self):
        """Test the _write_cell() method for formulas."""

        cell_tuple = namedtuple("Formula", "formula, format, value")
        cell = cell_tuple("A3+A5", None, 0)

        self.worksheet._write_cell(1, 2, cell)

        exp = """<c r="C2"><f>A3+A5</f><v>0</v></c>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_cell_formula02(self):
        """Test the _write_cell() method for formulas."""

        cell_tuple = namedtuple("Formula", "formula, format, value")
        cell = cell_tuple("A3+A5", None, 7)

        self.worksheet._write_cell(1, 2, cell)

        exp = """<c r="C2"><f>A3+A5</f><v>7</v></c>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("autofit01.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)
        worksheet = workbook.add_worksheet()

        # Before writing data, nothing to autofit (should not raise)
        worksheet.autofit()

        # Write something that can be autofit
        worksheet.write_string(0, 0, "A")

        # Check for handling default/None width.
        worksheet.set_column("A:A", None)
        worksheet.autofit()

        # Check for handling 0 width.
        worksheet.set_column("A:A", 0)
        worksheet.autofit()

        # Check for handling user defined width. Autofit shouldn't override.
        worksheet.set_column("A:A", 1.57143)
        worksheet.autofit()

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWriteSheetViews(unittest.TestCase):
    """
    Test the Worksheet _write_sheet_views() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_sheet_views1(self):
        """Test the _write_sheet_views() method with selection set"""

        self.worksheet.select()

        self.worksheet.set_selection("A1")

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"/></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views2(self):
        """Test the _write_sheet_views() method with selection set"""

        self.worksheet.select()

        self.worksheet.set_selection("A2")

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><selection activeCell="A2" sqref="A2"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views3(self):
        """Test the _write_sheet_views() method with selection set"""

        self.worksheet.select()

        self.worksheet.set_selection("B1")

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><selection activeCell="B1" sqref="B1"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views4(self):
        """Test the _write_sheet_views() method with selection set"""

        self.worksheet.select()

        self.worksheet.set_selection("D3")

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><selection activeCell="D3" sqref="D3"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views5(self):
        """Test the _write_sheet_views() method with selection set"""

        self.worksheet.select()

        self.worksheet.set_selection("D3:F4")

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><selection activeCell="D3" sqref="D3:F4"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views6(self):
        """Test the _write_sheet_views() method with selection set"""

        self.worksheet.select()

        # With reversed selection direction.
        self.worksheet.set_selection("F4:D3")

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><selection activeCell="F4" sqref="D3:F4"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_sheet_views7(self):
        """Test the _write_sheet_views() method with selection set"""

        self.worksheet.select()

        # Should be the same as 'A2'
        self.worksheet.set_selection("A2:A2")

        self.worksheet._write_sheet_views()

        exp = '<sheetViews><sheetView tabSelected="1" workbookViewId="0"><selection activeCell="A2" sqref="A2"/></sheetView></sheetViews>'
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("chart_order01.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet1 = workbook.add_worksheet()
        worksheet2 = workbook.add_worksheet()
        worksheet3 = workbook.add_worksheet()

        chart4 = workbook.add_chart({"type": "pie"})
        chart3 = workbook.add_chart({"type": "line"})
        chart2 = workbook.add_chart({"type": "bar"})
        chart1 = workbook.add_chart({"type": "column"})

        chart1.axis_ids = [54976896, 54978432]
        chart2.axis_ids = [54310784, 54312320]
        chart3.axis_ids = [69816704, 69818240]
        chart4.axis_ids = [69816704, 69818240]

        data = [
            [1, 2, 3, 4, 5],
            [2, 4, 6, 8, 10],
            [3, 6, 9, 12, 15],
        ]

        worksheet1.write_column("A1", data[0])
        worksheet1.write_column("B1", data[1])
        worksheet1.write_column("C1", data[2])

        worksheet2.write_column("A1", data[0])
        worksheet2.write_column("B1", data[1])
        worksheet2.write_column("C1", data[2])

        worksheet3.write_column("A1", data[0])
        worksheet3.write_column("B1", data[1])
        worksheet3.write_column("C1", data[2])

        chart1.add_series({"values": "=Sheet1!$A$1:$A$5"})
        chart2.add_series({"values": "=Sheet2!$A$1:$A$5"})
        chart3.add_series({"values": "=Sheet3!$A$1:$A$5"})
        chart4.add_series({"values": "=Sheet1!$B$1:$B$5"})

        worksheet1.insert_chart("E9", chart1)
        worksheet2.insert_chart("E9", chart2)
        worksheet3.insert_chart("E9", chart3)
        worksheet1.insert_chart("E24", chart4)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("set_row01.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.set_row(0, 0.75)
        worksheet.set_row(1, 1.50)
        worksheet.set_row(2, 2.25)
        worksheet.set_row(3, 3)

        worksheet.set_row(11, 9)
        worksheet.set_row(12, 9.75)
        worksheet.set_row(13, 10.50)
        worksheet.set_row(14, 11.25)

        worksheet.set_row(18, 14.25)
        worksheet.set_row(20, 15.75)
        worksheet.set_row(21, 16.50)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#
from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("defined_name04.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with defined names."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        # Test for valid Excel defined names.
        workbook.define_name("\\__", "=Sheet1!$A$1")
        workbook.define_name("a3f6", "=Sheet1!$A$2")
        workbook.define_name("afoo.bar", "=Sheet1!$A$3")
        workbook.define_name("étude", "=Sheet1!$A$4")
        workbook.define_name("eésumé", "=Sheet1!$A$5")
        workbook.define_name("b", "=Sheet1!$A$6")

        # The following aren't valid Excel names and shouldn't be written to
        # the output file. We ignore the warnings raised in define_name() and
        # instead check that the output file only contains the valid names.
        import warnings

        warnings.filterwarnings("ignore")

        workbook.define_name(".abc", "=Sheet1!$B$1")
        workbook.define_name("GFG$", "=Sheet1!$B$1")
        workbook.define_name("A1", "=Sheet1!$B$1")
        workbook.define_name("XFD1048576", "=Sheet1!$B$1")
        workbook.define_name("1A", "=Sheet1!$B$1")
        workbook.define_name("A A", "=Sheet1!$B$1")
        workbook.define_name("c", "=Sheet1!$B$1")
        workbook.define_name("r", "=Sheet1!$B$1")
        workbook.define_name("C", "=Sheet1!$B$1")
        workbook.define_name("R", "=Sheet1!$B$1")
        workbook.define_name("R1", "=Sheet1!$B$1")
        workbook.define_name("C1", "=Sheet1!$B$1")
        workbook.define_name("R1C1", "=Sheet1!$B$1")
        workbook.define_name("R13C99", "=Sheet1!$B$1")

        warnings.resetwarnings()

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("cond_format18.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with conditionalFormatting."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.write("A1", 1)
        worksheet.write("A2", 2)
        worksheet.write("A3", 3)
        worksheet.write("A4", 4)
        worksheet.write("A5", 5)
        worksheet.write("A6", 6)
        worksheet.write("A7", 7)
        worksheet.write("A8", 8)
        worksheet.write("A9", 9)

        worksheet.write("A12", 75)

        worksheet.conditional_format(
            "A1", {"type": "icon_set", "icon_style": "3_arrows", "reverse_icons": True}
        )

        worksheet.conditional_format(
            "A2", {"type": "icon_set", "icon_style": "3_flags", "icons_only": True}
        )

        worksheet.conditional_format(
            "A3",
            {
                "type": "icon_set",
                "icon_style": "3_traffic_lights_rimmed",
                "icons_only": True,
                "reverse_icons": True,
            },
        )

        worksheet.conditional_format(
            "A4",
            {
                "type": "icon_set",
                "icon_style": "3_symbols_circled",
                "icons": [{"value": 80}, {"value": 20}],
            },
        )

        worksheet.conditional_format(
            "A5",
            {
                "type": "icon_set",
                "icon_style": "4_arrows",
                "icons": [{"criteria": ">"}, {"criteria": ">"}, {"criteria": ">"}],
            },
        )

        worksheet.conditional_format(
            "A6",
            {
                "type": "icon_set",
                "icon_style": "4_red_to_black",
                "icons": [
                    {"criteria": ">=", "type": "number", "value": 90},
                    {"criteria": "<", "type": "percentile", "value": 50},
                    {"criteria": "<=", "type": "percent", "value": 25},
                ],
            },
        )

        worksheet.conditional_format(
            "A7",
            {
                "type": "icon_set",
                "icon_style": "4_traffic_lights",
                "icons": [{"value": "=$A$12"}],
            },
        )

        worksheet.conditional_format(
            "A8",
            {
                "type": "icon_set",
                "icon_style": "5_arrows_gray",
                "icons": [{"type": "formula", "value": "=$A$12"}],
            },
        )

        worksheet.conditional_format(
            "A9",
            {
                "type": "icon_set",
                "icon_style": "5_quarters",
                "icons": [{"value": 70}, {"value": 50}, {"value": 30}, {"value": 10}],
                "reverse_icons": True,
            },
        )

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWriteHyperlink(unittest.TestCase):
    """
    Test the Worksheet _write_hyperlink() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_hyperlink_external(self):
        """Test the _write_hyperlink() method"""

        self.worksheet._write_hyperlink_external(0, 0, 1)

        exp = """<hyperlink ref="A1" r:id="rId1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_hyperlink_internal_1(self):
        """Test the _write_hyperlink() method"""

        self.worksheet._write_hyperlink_internal(0, 0, "Sheet2!A1", "Sheet2!A1")

        exp = """<hyperlink ref="A1" location="Sheet2!A1" display="Sheet2!A1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_hyperlink_internal_2(self):
        """Test the _write_hyperlink() method"""

        self.worksheet._write_hyperlink_internal(
            4, 0, "'Data Sheet'!D5", "'Data Sheet'!D5"
        )

        exp = """<hyperlink ref="A5" location="'Data Sheet'!D5" display="'Data Sheet'!D5"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_hyperlink_internal_3(self):
        """Test the _write_hyperlink() method"""

        self.worksheet._write_hyperlink_internal(
            17, 0, "Sheet2!A1", "Sheet2!A1", "Screen Tip 1"
        )

        exp = """<hyperlink ref="A18" location="Sheet2!A1" tooltip="Screen Tip 1" display="Sheet2!A1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...sharedstrings import SharedStringTable
from ...sharedstrings import SharedStrings


class TestWriteSst(unittest.TestCase):
    """
    Test the SharedStrings _write_sst() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.sharedstrings = SharedStrings()
        self.sharedstrings._set_filehandle(self.fh)

    def test_write_sst(self):
        """Test the _write_sst() method"""

        string_table = SharedStringTable()

        # Add some strings and check the returned indices.
        string_table._get_shared_string_index("neptune")
        string_table._get_shared_string_index("neptune")
        string_table._get_shared_string_index("neptune")
        string_table._get_shared_string_index("mars")
        string_table._get_shared_string_index("venus")
        string_table._get_shared_string_index("mars")
        string_table._get_shared_string_index("venus")
        self.sharedstrings.string_table = string_table

        self.sharedstrings._write_sst()

        exp = """<sst xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" count="7" uniqueCount="3">"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("dynamic_array03.xlsx")

    def test_future_function01(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.write_formula("A1", "=1+_xlfn.XOR(1)", None, 2)

        workbook.close()

        self.assertExcelEqual()

    def test_future_function02(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()
        worksheet.use_future_functions = True

        worksheet.write_formula("A1", "=1+XOR(1)", None, 2)

        workbook.close()

        self.assertExcelEqual()

    def test_future_function03(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename, {"use_future_functions": True})

        worksheet = workbook.add_worksheet()

        worksheet.write_formula("A1", "=1+XOR(1)", None, 2)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("chart_bar14.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet1 = workbook.add_worksheet()
        worksheet2 = workbook.add_worksheet()
        worksheet3 = workbook.add_worksheet()
        chartsheet1 = workbook.add_chartsheet()
        chart1 = workbook.add_chart({"type": "bar"})
        chart2 = workbook.add_chart({"type": "bar"})
        chart3 = workbook.add_chart({"type": "column"})

        chart1.axis_ids = [40294272, 40295808]
        chart2.axis_ids = [40261504, 65749760]
        chart3.axis_ids = [65465728, 66388352]

        data = [
            [1, 2, 3, 4, 5],
            [2, 4, 6, 8, 10],
            [3, 6, 9, 12, 15],
        ]

        worksheet2.default_url_format = None

        worksheet2.write_column("A1", data[0])
        worksheet2.write_column("B1", data[1])
        worksheet2.write_column("C1", data[2])

        worksheet2.write("A6", "http://www.perl.com/")

        chart3.add_series({"values": "=Sheet2!$A$1:$A$5"})
        chart3.add_series({"values": "=Sheet2!$B$1:$B$5"})
        chart3.add_series({"values": "=Sheet2!$C$1:$C$5"})

        chart1.add_series({"values": "=Sheet2!$A$1:$A$5"})
        chart1.add_series({"values": "=Sheet2!$B$1:$B$5"})
        chart1.add_series({"values": "=Sheet2!$C$1:$C$5"})

        chart2.add_series({"values": "=Sheet2!$A$1:$A$5"})

        worksheet2.insert_chart("E9", chart1)
        worksheet2.insert_chart("F25", chart2)

        chartsheet1.set_chart(chart3)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("header_image09.xlsx")

        self.ignore_elements = {
            "xl/worksheets/sheet1.xml": ["<pageMargins", "<pageSetup"],
            "xl/worksheets/sheet2.xml": ["<pageMargins", "<pageSetup"],
        }

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with image(s)."""

        workbook = Workbook(self.got_filename)

        worksheet1 = workbook.add_worksheet()
        worksheet2 = workbook.add_worksheet()

        worksheet1.write("A1", "Foo")
        worksheet1.write_comment("B2", "Some text")

        worksheet1.set_comments_author("John")

        worksheet2.set_header("&L&G", {"image_left": self.image_dir + "red.jpg"})

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("cond_format12.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with conditional formatting."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        format1 = workbook.add_format(
            {"bg_color": "#FFFF00", "fg_color": "#FF0000", "pattern": 12}
        )

        worksheet.write("A1", "Hello", format1)

        worksheet.write("B3", 10)
        worksheet.write("B4", 20)
        worksheet.write("B5", 30)
        worksheet.write("B6", 40)

        worksheet.conditional_format(
            "B3:B6",
            {
                "type": "cell",
                "format": format1,
                "criteria": "greater than",
                "value": 20,
            },
        )

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook
from io import BytesIO


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("header_image04.xlsx")

        self.ignore_elements = {
            "xl/worksheets/sheet1.xml": ["<pageMargins", "<pageSetup"]
        }

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with image(s)."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.set_footer(
            "&L&G&C&G&R&G",
            {
                "image_left": self.image_dir + "red.jpg",
                "image_center": self.image_dir + "blue.jpg",
                "image_right": self.image_dir + "yellow.jpg",
            },
        )

        workbook.close()

        self.assertExcelEqual()

    def test_create_file_with_picture(self):
        """Test the creation of a simple XlsxWriter file with image(s)."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.set_footer(
            "&L&[Picture]&C&G&R&[Picture]",
            {
                "image_left": self.image_dir + "red.jpg",
                "image_center": self.image_dir + "blue.jpg",
                "image_right": self.image_dir + "yellow.jpg",
            },
        )

        workbook.close()

        self.assertExcelEqual()

    def test_create_file_from_bytesio(self):
        """Test the creation of a simple XlsxWriter file with image(s)."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        image_file_left = open(self.image_dir + "red.jpg", "rb")
        image_data_left = BytesIO(image_file_left.read())
        image_file_left.close()

        image_file_center = open(self.image_dir + "blue.jpg", "rb")
        image_data_center = BytesIO(image_file_center.read())
        image_file_center.close()

        image_file_right = open(self.image_dir + "yellow.jpg", "rb")
        image_data_right = BytesIO(image_file_right.read())
        image_file_right.close()

        worksheet.set_footer(
            "&L&G&C&G&R&G",
            {
                "image_left": "red.jpg",
                "image_center": "blue.jpg",
                "image_right": "yellow.jpg",
                "image_data_left": image_data_left,
                "image_data_center": image_data_center,
                "image_data_right": image_data_right,
            },
        )

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWritePageSetup(unittest.TestCase):
    """
    Test the Worksheet _write_page_setup() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_page_setup_none(self):
        """Test the _write_page_setup() method. Without any page setup"""

        self.worksheet._write_page_setup()

        exp = ""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_page_setup_landscape(self):
        """Test the _write_page_setup() method. With set_landscape()"""

        self.worksheet.set_landscape()

        self.worksheet._write_page_setup()

        exp = """<pageSetup orientation="landscape"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_page_setup_portrait(self):
        """Test the _write_page_setup() method. With set_portrait()"""

        self.worksheet.set_portrait()

        self.worksheet._write_page_setup()

        exp = """<pageSetup orientation="portrait"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_page_setup_paper(self):
        """Test the _write_page_setup() method. With set_paper()"""

        self.worksheet.set_paper(9)

        self.worksheet._write_page_setup()

        exp = """<pageSetup paperSize="9" orientation="portrait"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_page_setup_print_across(self):
        """Test the _write_page_setup() method. With print_across()"""

        self.worksheet.print_across()

        self.worksheet._write_page_setup()

        exp = """<pageSetup pageOrder="overThenDown" orientation="portrait"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("types02.xlsx")

    def test_write_boolean(self):
        """Test writing boolean."""

        workbook = Workbook(self.got_filename)
        worksheet = workbook.add_worksheet()

        worksheet.write_boolean(0, 0, True)
        worksheet.write_boolean(1, 0, False)

        workbook.close()

        self.assertExcelEqual()

    def test_write_boolean_write(self):
        """Test writing boolean with write()."""

        workbook = Workbook(self.got_filename)
        worksheet = workbook.add_worksheet()

        worksheet.write(0, 0, True)
        worksheet.write(1, 0, False)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ..helperfunctions import _xml_to_list
from ...styles import Styles
from ...workbook import Workbook


class TestAssembleStyles(unittest.TestCase):
    """
    Test assembling a complete Styles file.

    """

    def test_assemble_xml_file(self):
        """Test for number format and font styles."""
        self.maxDiff = None

        fh = StringIO()
        style = Styles()
        style._set_filehandle(fh)

        workbook = Workbook()

        workbook.add_format({"num_format": 2})
        workbook.add_format({"num_format": 2, "bold": 1})
        workbook.add_format({"num_format": "0.0"})

        workbook._set_default_xf_indices()
        workbook._prepare_format_properties()

        style._set_style_properties(
            [
                workbook.xf_formats,
                workbook.palette,
                workbook.font_count,
                workbook.num_formats,
                workbook.border_count,
                workbook.fill_count,
                workbook.custom_colors,
                workbook.dxf_formats,
                workbook.has_comments,
            ]
        )

        style._assemble_xml_file()
        workbook.fileclosed = 1

        exp = _xml_to_list(
            """
                <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <styleSheet xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main">
                  <numFmts count="1">
                    <numFmt numFmtId="164" formatCode="0.0"/>
                  </numFmts>
                  <fonts count="2">
                    <font>
                      <sz val="11"/>
                      <color theme="1"/>
                      <name val="Calibri"/>
                      <family val="2"/>
                      <scheme val="minor"/>
                    </font>
                    <font>
                      <b/>
                      <sz val="11"/>
                      <color theme="1"/>
                      <name val="Calibri"/>
                      <family val="2"/>
                      <scheme val="minor"/>
                    </font>
                  </fonts>
                  <fills count="2">
                    <fill>
                      <patternFill patternType="none"/>
                    </fill>
                    <fill>
                      <patternFill patternType="gray125"/>
                    </fill>
                  </fills>
                  <borders count="1">
                    <border>
                      <left/>
                      <right/>
                      <top/>
                      <bottom/>
                      <diagonal/>
                    </border>
                  </borders>
                  <cellStyleXfs count="1">
                    <xf numFmtId="0" fontId="0" fillId="0" borderId="0"/>
                  </cellStyleXfs>
                  <cellXfs count="4">
                    <xf numFmtId="0" fontId="0" fillId="0" borderId="0" xfId="0"/>
                    <xf numFmtId="2" fontId="0" fillId="0" borderId="0" xfId="0" applyNumberFormat="1"/>
                    <xf numFmtId="2" fontId="1" fillId="0" borderId="0" xfId="0" applyNumberFormat="1" applyFont="1"/>
                    <xf numFmtId="164" fontId="0" fillId="0" borderId="0" xfId="0" applyNumberFormat="1"/>
                  </cellXfs>
                  <cellStyles count="1">
                    <cellStyle name="Normal" xfId="0" builtinId="0"/>
                  </cellStyles>
                  <dxfs count="0"/>
                  <tableStyles count="0" defaultTableStyle="TableStyleMedium9" defaultPivotStyle="PivotStyleLight16"/>
                </styleSheet>
                """
        )

        got = _xml_to_list(fh.getvalue())

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from datetime import datetime
from ..helperfunctions import _xml_to_list
from ...core import Core


class TestAssembleCore(unittest.TestCase):
    """
    Test assembling a complete Core file.

    """

    def test_assemble_xml_file(self):
        """Test writing an Core file."""
        self.maxDiff = None

        fh = StringIO()
        core = Core()
        core._set_filehandle(fh)

        properties = {
            "author": "A User",
            "created": datetime(2010, 1, 1, 0, 0, 0),
        }

        core._set_properties(properties)

        core._assemble_xml_file()

        exp = _xml_to_list(
            """
                <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <cp:coreProperties xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:dcmitype="http://purl.org/dc/dcmitype/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                  <dc:creator>A User</dc:creator>
                  <cp:lastModifiedBy>A User</cp:lastModifiedBy>
                  <dcterms:created xsi:type="dcterms:W3CDTF">2010-01-01T00:00:00Z</dcterms:created>
                  <dcterms:modified xsi:type="dcterms:W3CDTF">2010-01-01T00:00:00Z</dcterms:modified>
                </cp:coreProperties>
                """
        )

        got = _xml_to_list(fh.getvalue())

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...sharedstrings import SharedStrings


class TestWriteSi(unittest.TestCase):
    """
    Test the SharedStrings _write_si() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.sharedstrings = SharedStrings()
        self.sharedstrings._set_filehandle(self.fh)

    def test_write_si(self):
        """Test the _write_si() method"""

        self.sharedstrings._write_si("neptune")

        exp = """<si><t>neptune</t></si>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ..helperfunctions import _xml_to_list, _vml_to_list
from ...vml import Vml


class TestAssembleVml(unittest.TestCase):
    """
    Test assembling a complete Vml file.

    """

    def test_assemble_xml_file(self):
        """Test writing a vml with no cell data."""
        self.maxDiff = None

        fh = StringIO()
        vml = Vml()
        vml._set_filehandle(fh)

        vml._assemble_xml_file(
            1,
            1024,
            [
                [
                    1,
                    1,
                    "Some text",
                    "",
                    None,
                    "#ffffe1",
                    "Tahoma",
                    8,
                    2,
                    [2, 0, 15, 10, 4, 4, 15, 4, 143, 10, 128, 74],
                ]
            ],
            [],
        )

        exp = _vml_to_list(
            """
                <xml xmlns:v="urn:schemas-microsoft-com:vml"
                 xmlns:o="urn:schemas-microsoft-com:office:office"
                 xmlns:x="urn:schemas-microsoft-com:office:excel">
                 <o:shapelayout v:ext="edit">
                  <o:idmap v:ext="edit" data="1"/>
                 </o:shapelayout><v:shapetype id="_x0000_t202" coordsize="21600,21600" o:spt="202"
                  path="m,l,21600r21600,l21600,xe">
                  <v:stroke joinstyle="miter"/>
                  <v:path gradientshapeok="t" o:connecttype="rect"/>
                 </v:shapetype><v:shape id="_x0000_s1025" type="#_x0000_t202" style='position:absolute;
                  margin-left:107.25pt;margin-top:7.5pt;width:96pt;height:55.5pt;z-index:1;
                  visibility:hidden' fillcolor="#ffffe1" o:insetmode="auto">
                  <v:fill color2="#ffffe1"/>
                  <v:shadow on="t" color="black" obscured="t"/>
                  <v:path o:connecttype="none"/>
                  <v:textbox style='mso-direction-alt:auto'>
                   <div style='text-align:left'></div>
                  </v:textbox>
                  <x:ClientData ObjectType="Note">
                   <x:MoveWithCells/>
                   <x:SizeWithCells/>
                   <x:Anchor>
                    2, 15, 0, 10, 4, 15, 4, 4</x:Anchor>
                   <x:AutoFill>False</x:AutoFill>
                   <x:Row>1</x:Row>
                   <x:Column>1</x:Column>
                  </x:ClientData>
                 </v:shape></xml>
                """
        )

        got = _xml_to_list(fh.getvalue())

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook
import uuid


def write_uuid(worksheet, row, col, token, format=None):
    return worksheet.write_string(row, col, str(token), format)


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("types10.xlsx")

    def test_write_user_type(self):
        """Test writing numbers as text."""

        workbook = Workbook(self.got_filename)
        worksheet = workbook.add_worksheet()

        worksheet.add_write_handler(uuid.UUID, write_uuid)
        my_uuid = uuid.uuid3(uuid.NAMESPACE_DNS, "python.org")

        worksheet.write("A1", my_uuid)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from ...utility import xl_cell_to_rowcol_abs


class TestUtility(unittest.TestCase):
    """
    Test xl_cell_to_rowcol_abs() utility function.

    """

    def test_xl_cell_to_rowcol_abs(self):
        """Test xl_cell_to_rowcol_abs()"""

        tests = [
            # row, col, A1 string
            (0, 0, "A1"),
            (0, 1, "B1"),
            (0, 2, "C1"),
            (0, 9, "J1"),
            (1, 0, "A2"),
            (2, 0, "A3"),
            (9, 0, "A10"),
            (1, 24, "Y2"),
            (7, 25, "Z8"),
            (9, 26, "AA10"),
            (1, 254, "IU2"),
            (1, 255, "IV2"),
            (1, 256, "IW2"),
            (0, 16383, "XFD1"),
            (1048576, 16384, "XFE1048577"),
        ]

        for row, col, string in tests:
            exp = (row, col, 0, 0)
            got = xl_cell_to_rowcol_abs(string)
            self.assertEqual(got, exp)

    def test_xl_cell_to_rowcol_abs_abs(self):
        """Test xl_cell_to_rowcol_abs() with absolute references"""

        tests = [
            # row, col, row_abs, col_abs, A1 string
            (0, 0, 0, 0, "A1"),
            (0, 0, 1, 0, "A$1"),
            (0, 0, 0, 1, "$A1"),
            (0, 0, 1, 1, "$A$1"),
        ]

        for row, col, row_abs, col_abs, string in tests:
            exp = (row, col, row_abs, col_abs)
            got = xl_cell_to_rowcol_abs(string)
            self.assertEqual(got, exp)



##############################################################################
#
# Simple Python program to benchmark several Python Excel writing modules.
#
# python bench_excel_writers.py [num_rows] [num_cols]
#
# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
#

import sys
from time import perf_counter

import openpyxl
import pyexcelerate
import xlsxwriter
import xlwt


# Default to 1000 rows x 50 cols.
if len(sys.argv) > 1:
    row_max = int(sys.argv[1])
    col_max = 50
else:
    row_max = 1000
    col_max = 50

if len(sys.argv) > 2:
    col_max = int(sys.argv[2])


def print_elapsed_time(module_name, elapsed):
    """ Print module run times in a consistent format. """
    print("    %-28s: %6.2f" % (module_name, elapsed))


def time_xlsxwriter():
    """ Run XlsxWriter in default mode. """
    start_time = perf_counter()

    workbook = xlsxwriter.Workbook('xlsxwriter.xlsx')
    worksheet = workbook.add_worksheet()

    for row in range(row_max // 2):
        for col in range(col_max):
            worksheet.write_string(row * 2, col, "Row: %d Col: %d" % (row, col))
        for col in range(col_max):
            worksheet.write_number(row * 2 + 1, col, row + col)

    workbook.close()

    elapsed = perf_counter() - start_time
    print_elapsed_time('xlsxwriter', elapsed)


def time_xlsxwriter_optimised():
    """ Run XlsxWriter in optimised/constant memory mode. """
    start_time = perf_counter()

    workbook = xlsxwriter.Workbook('xlsxwriter_opt.xlsx',
                                   {'constant_memory': True})
    worksheet = workbook.add_worksheet()

    for row in range(row_max // 2):
        for col in range(col_max):
            worksheet.write_string(row * 2, col, "Row: %d Col: %d" % (row, col))
        for col in range(col_max):
            worksheet.write_number(row * 2 + 1, col, row + col)

    workbook.close()

    elapsed = perf_counter() - start_time
    print_elapsed_time('xlsxwriter (constant_memory)', elapsed)


def time_openpyxl():
    """ Run OpenPyXL in default mode. """
    start_time = perf_counter()

    workbook = openpyxl.workbook.Workbook()
    worksheet = workbook.active

    for row in range(row_max // 2):
        for col in range(col_max):
            worksheet.cell(row * 2 + 1, col + 1, "Row: %d Col: %d" % (row, col))
        for col in range(col_max):
            worksheet.cell(row * 2 + 2, col + 1, row + col)

    workbook.save('openpyxl.xlsx')

    elapsed = perf_counter() - start_time
    print_elapsed_time('openpyxl', elapsed)


def time_openpyxl_optimised():
    """ Run OpenPyXL in optimised mode. """
    start_time = perf_counter()

    workbook = openpyxl.workbook.Workbook(write_only=True)
    worksheet = workbook.create_sheet()

    for row in range(row_max // 2):
        string_data = ["Row: %d Col: %d" % (row, col) for col in range(col_max)]
        worksheet.append(string_data)

        num_data = [row + col for col in range(col_max)]
        worksheet.append(num_data)

    workbook.save('openpyxl_opt.xlsx')

    elapsed = perf_counter() - start_time
    print_elapsed_time('openpyxl   (optimised)', elapsed)


def time_pyexcelerate():
    """ Run pyexcelerate in "faster" mode. """
    start_time = perf_counter()

    workbook = pyexcelerate.Workbook()
    worksheet = workbook.new_sheet('Sheet1')

    for row in range(row_max // 2):
        for col in range(col_max):
            worksheet.set_cell_value(row * 2 + 1, col + 1, "Row: %d Col: %d" % (row, col))
        for col in range(col_max):
            worksheet.set_cell_value(row * 2 + 2, col + 1, row + col)

    workbook.save('pyexcelerate.xlsx')
    elapsed = perf_counter() - start_time

    print_elapsed_time('pyexcelerate', elapsed)


def time_xlwt():
    """ Run xlwt in default mode. """
    start_time = perf_counter()

    workbook = xlwt.Workbook()
    worksheet = workbook.add_sheet('Sheet1')

    for row in range(row_max // 2):
        for col in range(col_max):
            worksheet.write(row * 2, col, "Row: %d Col: %d" % (row, col))
        for col in range(col_max):
            worksheet.write(row * 2 + 1, col, row + col)

    workbook.save('xlwt.xls')

    elapsed = perf_counter() - start_time
    print_elapsed_time('xlwt', elapsed)


print("")
print("Versions:")
print("    %-12s: %s" % ('python', str(sys.version).split()[0]))
print("    %-12s: %s" % ('openpyxl', openpyxl.__version__))
print("    %-12s: %s" % ('pyexcelerate', pyexcelerate.__version__))
print("    %-12s: %s" % ('xlsxwriter', xlsxwriter.__version__))
print("    %-12s: %s" % ('xlwt', xlwt.__VERSION__))
print("")

print("Dimensions:")
print("    Rows = %d" % row_max)
print("    Cols = %d" % col_max)
print("")

print("Times:")
time_pyexcelerate()
time_xlwt()
time_xlsxwriter_optimised()
time_xlsxwriter()
time_openpyxl_optimised()
time_openpyxl()
print("")



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("types05.xlsx")

        self.ignore_files = [
            "xl/calcChain.xml",
            "[Content_Types].xml",
            "xl/_rels/workbook.xml.rels",
        ]

    def test_write_formula_default(self):
        """Test writing formulas with strings_to_formulas on."""

        workbook = Workbook(self.got_filename)
        worksheet = workbook.add_worksheet()

        worksheet.write(0, 0, "=1+1", None, 2)
        worksheet.write_string(1, 0, "=1+1")

        workbook.close()

        self.assertExcelEqual()

    def test_write_formula_implicit(self):
        """Test writing formulas with strings_to_formulas on."""

        workbook = Workbook(self.got_filename, {"strings_to_formulas": True})
        worksheet = workbook.add_worksheet()

        worksheet.write(0, 0, "=1+1", None, 2)
        worksheet.write_string(1, 0, "=1+1")

        workbook.close()

        self.assertExcelEqual()

    def test_write_formula_explicit(self):
        """Test writing formulas with strings_to_formulas off."""

        workbook = Workbook(self.got_filename, {"strings_to_formulas": False})
        worksheet = workbook.add_worksheet()

        worksheet.write_formula(0, 0, "=1+1", None, 2)
        worksheet.write(1, 0, "=1+1")

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("chart_clustered01.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()
        chart = workbook.add_chart({"type": "column"})

        chart.axis_ids = [45886080, 45928832]

        data = [
            ["Types", "Sub Type", "Value 1", "Value 2", "Value 3"],
            ["Type 1", "Sub Type A", 5000, 8000, 6000],
            ["", "Sub Type B", 2000, 3000, 4000],
            ["", "Sub Type C", 250, 1000, 2000],
            ["Type 2", "Sub Type D", 6000, 6000, 6500],
            ["", "Sub Type E", 500, 300, 200],
        ]

        cat_data = [
            ["Type 1", None, None, "Type 2", None],
            ["Sub Type A", "Sub Type B", "Sub Type C", "Sub Type D", "Sub Type E"],
        ]

        for row_num, row_data in enumerate(data):
            worksheet.write_row(row_num, 0, row_data)

        chart.add_series(
            {
                "name": "=Sheet1!$C$1",
                "categories": "=Sheet1!$A$2:$B$6",
                "values": "=Sheet1!$C$2:$C$6",
                "categories_data": cat_data,
            }
        )

        chart.add_series(
            {
                "name": "=Sheet1!$D$1",
                "categories": "=Sheet1!$A$2:$B$6",
                "values": "=Sheet1!$D$2:$D$6",
            }
        )

        chart.add_series(
            {
                "name": "=Sheet1!$E$1",
                "categories": "=Sheet1!$A$2:$B$6",
                "values": "=Sheet1!$E$2:$E$6",
            }
        )

        worksheet.insert_chart("E9", chart)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWritePrintOptions(unittest.TestCase):
    """
    Test the Worksheet _write_print_options() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_print_options_default(self):
        """Test the _write_print_options() method without options"""

        self.worksheet._write_print_options()

        exp = """"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_print_options_hcenter(self):
        """Test the _write_print_options() method with horizontal center"""

        self.worksheet.center_horizontally()
        self.worksheet._write_print_options()

        exp = """<printOptions horizontalCentered="1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_print_options_vcenter(self):
        """Test the _write_print_options() method with vertical center"""

        self.worksheet.center_vertically()
        self.worksheet._write_print_options()

        exp = """<printOptions verticalCentered="1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_print_options_center(self):
        """Test the _write_print_options() method with horiz + vert center"""

        self.worksheet.center_horizontally()
        self.worksheet.center_vertically()
        self.worksheet._write_print_options()

        exp = """<printOptions horizontalCentered="1" verticalCentered="1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_print_options_gridlines_default(self):
        """Test the _write_print_options() method with default value"""

        self.worksheet.hide_gridlines()
        self.worksheet._write_print_options()

        exp = """"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_print_options_gridlines_0(self):
        """Test the _write_print_options() method with 0 value"""

        self.worksheet.hide_gridlines(0)
        self.worksheet._write_print_options()

        exp = """<printOptions gridLines="1"/>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...workbook import Workbook


class TestAssembleWorkbook(unittest.TestCase):
    """
    Test assembling a complete Workbook file.

    """

    def test_get_worksheet_by_name01(self):
        """Test get_worksheet_by_name()"""
        fh = StringIO()
        workbook = Workbook()
        workbook._set_filehandle(fh)

        exp = workbook.add_worksheet()
        got = workbook.get_worksheet_by_name("Sheet1")
        workbook.fileclosed = 1

        self.assertEqual(got, exp)

    def test_get_worksheet_by_name02(self):
        """Test get_worksheet_by_name()"""
        fh = StringIO()
        workbook = Workbook()
        workbook._set_filehandle(fh)

        workbook.add_worksheet()
        exp = workbook.add_worksheet()
        got = workbook.get_worksheet_by_name("Sheet2")
        workbook.fileclosed = 1

        self.assertEqual(got, exp)

    def test_get_worksheet_by_name03(self):
        """Test get_worksheet_by_name()"""
        fh = StringIO()
        workbook = Workbook()
        workbook._set_filehandle(fh)

        exp = workbook.add_worksheet("Sheet 3")
        got = workbook.get_worksheet_by_name("Sheet 3")
        workbook.fileclosed = 1

        self.assertEqual(got, exp)

    def test_get_worksheet_by_name04(self):
        """Test get_worksheet_by_name()"""
        fh = StringIO()
        workbook = Workbook()
        workbook._set_filehandle(fh)

        exp = workbook.add_worksheet("Sheet '4")
        got = workbook.get_worksheet_by_name("Sheet '4")
        workbook.fileclosed = 1

        self.assertEqual(got, exp)

    def test_get_worksheet_by_name05(self):
        """Test get_worksheet_by_name()"""
        fh = StringIO()
        workbook = Workbook()
        workbook._set_filehandle(fh)

        exp = None
        got = workbook.get_worksheet_by_name("Sheet 5")
        workbook.fileclosed = 1

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook
from datetime import datetime


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("properties04.xlsx")

    def test_create_file_explicit(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        long_string = "This is a long string. " * 11
        long_string += "AA"

        date = datetime.strptime("2016-12-12 23:00:00", "%Y-%m-%d %H:%M:%S")

        # This test uses explicit property types.
        workbook.set_custom_property("Checked by", "Adam", "text")
        workbook.set_custom_property("Date completed", date, "date")
        workbook.set_custom_property("Document number", 12345, "number_int")
        workbook.set_custom_property("Reference", 1.2345, "number")
        workbook.set_custom_property("Source", True, "bool")
        workbook.set_custom_property("Status", False, "bool")
        workbook.set_custom_property("Department", long_string, "text")
        workbook.set_custom_property("Group", 1.23456789012, "number")

        worksheet.set_column("A:A", 70)
        worksheet.write(
            "A1",
            "Select 'Office Button -> Prepare -> Properties' to see the file properties.",
        )

        workbook.close()

        self.assertExcelEqual()

    def test_create_file_implicit(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        long_string = "This is a long string. " * 11
        long_string += "AA"

        date = datetime.strptime("2016-12-12 23:00:00", "%Y-%m-%d %H:%M:%S")

        # This test uses implicit property types.
        workbook.set_custom_property("Checked by", "Adam")
        workbook.set_custom_property("Date completed", date)
        workbook.set_custom_property("Document number", 12345)
        workbook.set_custom_property("Reference", 1.2345)
        workbook.set_custom_property("Source", True)
        workbook.set_custom_property("Status", False)
        workbook.set_custom_property("Department", long_string)
        workbook.set_custom_property("Group", 1.23456789012)

        worksheet.set_column("A:A", 70)
        worksheet.write(
            "A1",
            "Select 'Office Button -> Prepare -> Properties' to see the file properties.",
        )

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from datetime import date
from io import StringIO
from ...worksheet import Worksheet
from ..helperfunctions import _xml_to_list


class TestWriteDataValidations(unittest.TestCase):
    """
    Test the Worksheet _write_data_validations() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_data_validations_1(self):
        """Test the _write_data_validations() method. Data validation example 1 from docs"""

        self.worksheet.data_validation(
            "A1",
            {
                "validate": "integer",
                "criteria": ">",
                "value": 0,
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="whole" operator="greaterThan" allowBlank="1" showInputMessage="1" showErrorMessage="1" sqref="A1"><formula1>0</formula1></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)

    def test_write_data_validations_1b(self):
        """Test the _write_data_validations() method. Data validation example 1 from docs (with options turned off)"""

        self.worksheet.data_validation(
            "A1",
            {
                "validate": "integer",
                "criteria": ">",
                "value": 0,
                "ignore_blank": 0,
                "show_input": 0,
                "show_error": 0,
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="whole" operator="greaterThan" sqref="A1"><formula1>0</formula1></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)

    def test_write_data_validations_2(self):
        """Test the _write_data_validations() method. Data validation example 2 from docs"""

        self.worksheet.data_validation(
            "A2",
            {
                "validate": "integer",
                "criteria": ">",
                "value": "=E3",
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="whole" operator="greaterThan" allowBlank="1" showInputMessage="1" showErrorMessage="1" sqref="A2"><formula1>E3</formula1></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)

    def test_write_data_validations_3(self):
        """Test the _write_data_validations() method. Data validation example 3 from docs"""

        self.worksheet.data_validation(
            "A3",
            {
                "validate": "decimal",
                "criteria": "between",
                "minimum": 0.1,
                "maximum": 0.5,
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="decimal" allowBlank="1" showInputMessage="1" showErrorMessage="1" sqref="A3"><formula1>0.1</formula1><formula2>0.5</formula2></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)

    def test_write_data_validations_4(self):
        """Test the _write_data_validations() method. Data validation example 4 from docs"""

        self.worksheet.data_validation(
            "A4",
            {
                "validate": "list",
                "source": ["open", "high", "close"],
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="list" allowBlank="1" showInputMessage="1" showErrorMessage="1" sqref="A4"><formula1>"open,high,close"</formula1></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)

    def test_write_data_validations_5(self):
        """Test the _write_data_validations() method. Data validation example 5 from docs"""

        self.worksheet.data_validation(
            "A5",
            {
                "validate": "list",
                "source": "=$E$4:$G$4",
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="list" allowBlank="1" showInputMessage="1" showErrorMessage="1" sqref="A5"><formula1>$E$4:$G$4</formula1></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)

    def test_write_data_validations_6(self):
        """Test the _write_data_validations() method. Data validation example 6 from docs"""

        self.worksheet.data_validation(
            "A6",
            {
                "validate": "date",
                "criteria": "between",
                "minimum": date(2008, 1, 1),
                "maximum": date(2008, 12, 12),
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="date" allowBlank="1" showInputMessage="1" showErrorMessage="1" sqref="A6"><formula1>39448</formula1><formula2>39794</formula2></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)

    def test_write_data_validations_7(self):
        """Test the _write_data_validations() method. Data validation example 7 from docs"""

        self.worksheet.data_validation(
            "A7",
            {
                "validate": "integer",
                "criteria": "between",
                "minimum": 1,
                "maximum": 100,
                "input_title": "Enter an integer:",
                "input_message": "between 1 and 100",
            },
        )

        self.worksheet._write_data_validations()

        exp = """<dataValidations count="1"><dataValidation type="whole" allowBlank="1" showInputMessage="1" showErrorMessage="1" promptTitle="Enter an integer:" prompt="between 1 and 100" sqref="A7"><formula1>1</formula1><formula2>100</formula2></dataValidation></dataValidations>"""
        got = self.fh.getvalue()

        exp = _xml_to_list(exp)
        got = _xml_to_list(got)

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("types01.xlsx")

    def test_write_number_as_text(self):
        """Test writing numbers as text."""

        workbook = Workbook(self.got_filename)
        worksheet = workbook.add_worksheet()

        worksheet.write_string(0, 0, "Hello")
        worksheet.write_string(1, 0, "123")

        workbook.close()

        self.assertExcelEqual()

    def test_write_number_as_text_with_write(self):
        """Test writing numbers as text using write() without conversion."""

        workbook = Workbook(self.got_filename, {"strings_to_numbers": False})
        worksheet = workbook.add_worksheet()

        worksheet.write(0, 0, "Hello")
        worksheet.write(1, 0, "123")

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("simple05.xlsx")

    def test_create_file(self):
        """Test font formatting."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.set_row(5, 18)
        worksheet.set_row(6, 18)

        format1 = workbook.add_format({"bold": 1})
        format2 = workbook.add_format({"italic": 1})
        format3 = workbook.add_format({"bold": 1, "italic": 1})
        format4 = workbook.add_format({"underline": 1})
        format5 = workbook.add_format({"font_strikeout": 1})
        format6 = workbook.add_format({"font_script": 1})
        format7 = workbook.add_format({"font_script": 2})

        worksheet.write_string(0, 0, "Foo", format1)
        worksheet.write_string(1, 0, "Foo", format2)
        worksheet.write_string(2, 0, "Foo", format3)
        worksheet.write_string(3, 0, "Foo", format4)
        worksheet.write_string(4, 0, "Foo", format5)
        worksheet.write_string(5, 0, "Foo", format6)
        worksheet.write_string(6, 0, "Foo", format7)

        workbook.close()

        self.assertExcelEqual()

    def test_create_file_in_memory(self):
        """Test font formatting."""

        workbook = Workbook(self.got_filename, {"in_memory": True})

        worksheet = workbook.add_worksheet()

        worksheet.set_row(5, 18)
        worksheet.set_row(6, 18)

        format1 = workbook.add_format({"bold": 1})
        format2 = workbook.add_format({"italic": 1})
        format3 = workbook.add_format({"bold": 1, "italic": 1})
        format4 = workbook.add_format({"underline": 1})
        format5 = workbook.add_format({"font_strikeout": 1})
        format6 = workbook.add_format({"font_script": 1})
        format7 = workbook.add_format({"font_script": 2})

        worksheet.write_string(0, 0, "Foo", format1)
        worksheet.write_string(1, 0, "Foo", format2)
        worksheet.write_string(2, 0, "Foo", format3)
        worksheet.write_string(3, 0, "Foo", format4)
        worksheet.write_string(4, 0, "Foo", format5)
        worksheet.write_string(5, 0, "Foo", format6)
        worksheet.write_string(6, 0, "Foo", format7)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("date_examples01.xlsx")

    def test_create_file(self):
        """Example spreadsheet used in the tutorial."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        # Widen column A for extra visibility.
        worksheet.set_column("A:A", 30)

        # A number to convert to a date.
        number = 41333.5

        # Write it as a number without formatting.
        worksheet.write("A1", number)  # 41333.5

        format2 = workbook.add_format({"num_format": "dd/mm/yy"})
        worksheet.write("A2", number, format2)  # 28/02/13

        format3 = workbook.add_format({"num_format": "mm/dd/yy"})
        worksheet.write("A3", number, format3)  # 02/28/13

        format4 = workbook.add_format({"num_format": "d\\-m\\-yyyy"})
        worksheet.write("A4", number, format4)  # 28-2-2013

        format5 = workbook.add_format({"num_format": "dd/mm/yy\\ hh:mm"})
        worksheet.write("A5", number, format5)  # 28/02/13 12:00

        format6 = workbook.add_format({"num_format": "d\\ mmm\\ yyyy"})
        worksheet.write("A6", number, format6)  # 28 Feb 2013

        format7 = workbook.add_format({"num_format": "mmm\\ d\\ yyyy\\ hh:mm\\ AM/PM"})
        worksheet.write("A7", number, format7)  # Feb 28 2008 12:00 PM

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("shared_strings02.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        strings = [
            "_",
            "_x",
            "_x0",
            "_x00",
            "_x000",
            "_x0000",
            "_x0000_",
            "_x005F_",
            "_x000G_",
            "_X0000_",
            "_x000a_",
            "_x000A_",
            "_x0000__x0000_",
            "__x0000__",
        ]

        worksheet.write_column(0, 0, strings)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("protect06.xlsx")

    def test_create_file(self):
        """Test the a simple XlsxWriter file with worksheet protection."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        unlocked = workbook.add_format({"locked": 0, "hidden": 0})
        hidden = workbook.add_format({"locked": 0, "hidden": 1})

        worksheet.protect()

        worksheet.unprotect_range("A1", None, "password")
        worksheet.unprotect_range("C1:C3")
        worksheet.unprotect_range("G4:I6", "MyRange")
        worksheet.unprotect_range("K7", None, "foobar")

        worksheet.write("A1", 1)
        worksheet.write("A2", 2, unlocked)
        worksheet.write("A3", 3, hidden)

        workbook.close()

        self.assertExcelEqual()

    def test_create_file_keyword(self):
        """Test the a simple XlsxWriter file with worksheet protection."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        unlocked = workbook.add_format({"locked": 0, "hidden": 0})
        hidden = workbook.add_format({"locked": 0, "hidden": 1})

        worksheet.protect()

        worksheet.unprotect_range("A1", password="password")
        worksheet.unprotect_range(cell_range="C1:C3")
        worksheet.unprotect_range("G4:I6", range_name="MyRange")
        worksheet.unprotect_range("K7", None, "foobar")

        worksheet.write("A1", 1)
        worksheet.write("A2", 2, unlocked)
        worksheet.write("A3", 3, hidden)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from io import StringIO
from ...worksheet import Worksheet


class TestWriteHeaderFooter(unittest.TestCase):
    """
    Test the Worksheet _write_header_footer() method.

    """

    def setUp(self):
        self.fh = StringIO()
        self.worksheet = Worksheet()
        self.worksheet._set_filehandle(self.fh)

    def test_write_header_footer_header_only(self):
        """Test the _write_header_footer() method header only"""

        self.worksheet.set_header("Page &P of &N")
        self.worksheet._write_header_footer()

        exp = """<headerFooter><oddHeader>Page &amp;P of &amp;N</oddHeader></headerFooter>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_header_footer_footer_only(self):
        """Test the _write_header_footer() method footer only"""

        self.worksheet.set_footer("&F")
        self.worksheet._write_header_footer()

        exp = """<headerFooter><oddFooter>&amp;F</oddFooter></headerFooter>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)

    def test_write_header_footer_both(self):
        """Test the _write_header_footer() method header and footer"""

        self.worksheet.set_header("Page &P of &N")
        self.worksheet.set_footer("&F")
        self.worksheet._write_header_footer()

        exp = """<headerFooter><oddHeader>Page &amp;P of &amp;N</oddHeader><oddFooter>&amp;F</oddFooter></headerFooter>"""
        got = self.fh.getvalue()

        self.assertEqual(got, exp)



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

import unittest
from ...workbook import Workbook
from ...exceptions import DuplicateWorksheetName
from ...exceptions import InvalidWorksheetName


class TestCheckSheetname(unittest.TestCase):
    """
    Test the Workbook _check_sheetname() method.

    """

    def setUp(self):
        self.workbook = Workbook()

    def test_check_sheetname(self):
        """Test the _check_sheetname() method"""

        got = self.workbook._check_sheetname("name")
        exp = "name"
        self.assertEqual(got, exp)

        got = self.workbook._check_sheetname("Sheet1")
        exp = "Sheet1"
        self.assertEqual(got, exp)

        got = self.workbook._check_sheetname(None)
        exp = "Sheet3"
        self.assertEqual(got, exp)

        got = self.workbook._check_sheetname("")
        exp = "Sheet4"
        self.assertEqual(got, exp)

    def test_check_sheetname_with_long_name(self):
        """Test the _check_sheetname() method with exception"""

        name = "name_that_is_longer_than_thirty_one_characters"
        self.assertRaises(InvalidWorksheetName, self.workbook._check_sheetname, name)

    def test_check_sheetname_with_invalid_name(self):
        """Test the _check_sheetname() method with exception"""

        name = "name_with_special_character_?"
        self.assertRaises(InvalidWorksheetName, self.workbook._check_sheetname, name)

        name = "'start with apostrophe"
        self.assertRaises(InvalidWorksheetName, self.workbook._check_sheetname, name)

        name = "end with apostrophe'"
        self.assertRaises(InvalidWorksheetName, self.workbook._check_sheetname, name)

        name = "'start and end with apostrophe'"
        self.assertRaises(InvalidWorksheetName, self.workbook._check_sheetname, name)

    def test_check_sheetname_with_duplicate_name(self):
        """Test the _check_sheetname() method with exception"""

        name1 = "Duplicate_name"
        name2 = name1.lower()

        self.workbook.add_worksheet(name1)
        self.assertRaises(DuplicateWorksheetName, self.workbook.add_worksheet, name2)

    def tearDown(self):
        self.workbook.fileclosed = True



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("chart_display_units10.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()
        chart = workbook.add_chart({"type": "column"})

        chart.axis_ids = [56159232, 61364096]

        data = [
            [10000000, 20000000, 30000000, 20000000, 10000000],
        ]

        worksheet.write_column(0, 0, data[0])

        chart.add_series({"values": "=Sheet1!$A$1:$A$5"})

        chart.set_y_axis({"display_units": "trillions", "display_units_visible": 0})

        worksheet.insert_chart("E9", chart)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("cond_format04.xlsx")

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with conditional formatting."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        format1 = workbook.add_format({"num_format": 2, "dxf_index": 1})
        format2 = workbook.add_format({"num_format": "0.000", "dxf_index": 0})

        worksheet.write("A1", 10)
        worksheet.write("A2", 20)
        worksheet.write("A3", 30)
        worksheet.write("A4", 40)

        worksheet.conditional_format(
            "A1",
            {
                "type": "cell",
                "format": format1,
                "criteria": ">",
                "value": 2,
            },
        )

        worksheet.conditional_format(
            "A2",
            {
                "type": "cell",
                "format": format2,
                "criteria": "<",
                "value": 8,
            },
        )

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("chart_drop_lines04.xlsx")

        self.ignore_elements = {"xl/charts/chart1.xml": ["<c:formatCode"]}

    def test_create_file(self):
        """Test the creation of an XlsxWriter file with drop down lines."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()
        chart = workbook.add_chart({"type": "stock"})
        date_format = workbook.add_format({"num_format": 14})

        chart.axis_ids = [49019520, 49222016]

        data = [
            [39083, 39084, 39085, 39086, 39087],
            [27.2, 25.03, 19.05, 20.34, 18.5],
            [23.49, 19.55, 15.12, 17.84, 16.34],
            [25.45, 23.05, 17.32, 20.45, 17.34],
        ]

        for row in range(5):
            worksheet.write(row, 0, data[0][row], date_format)
            worksheet.write(row, 1, data[1][row])
            worksheet.write(row, 2, data[2][row])
            worksheet.write(row, 3, data[3][row])

        worksheet.set_column("A:D", 11)

        chart.add_series(
            {
                "categories": "=Sheet1!$A$1:$A$5",
                "values": "=Sheet1!$B$1:$B$5",
            }
        )

        chart.add_series(
            {
                "categories": "=Sheet1!$A$1:$A$5",
                "values": "=Sheet1!$C$1:$C$5",
            }
        )

        chart.add_series(
            {
                "categories": "=Sheet1!$A$1:$A$5",
                "values": "=Sheet1!$D$1:$D$5",
            }
        )

        chart.set_drop_lines()

        worksheet.insert_chart("E9", chart)

        workbook.close()

        self.assertExcelEqual()



###############################################################################
#
# Tests for XlsxWriter.
#
# SPDX-License-Identifier: BSD-2-Clause
# Copyright (c), 2013-2023, John McNamara, jmcnamara@cpan.org
#

from ..excel_comparison_test import ExcelComparisonTest
from ...workbook import Workbook
from ...sharedstrings import SharedStringTable


class TestCompareXLSXFiles(ExcelComparisonTest):
    """
    Test file created by XlsxWriter against a file created by Excel.

    """

    def setUp(self):
        self.set_filename("hyperlink19.xlsx")

        self.ignore_files = [
            "xl/calcChain.xml",
            "[Content_Types].xml",
            "xl/_rels/workbook.xml.rels",
        ]

    def test_create_file(self):
        """Test the creation of a simple XlsxWriter file with hyperlinks."""

        workbook = Workbook(self.got_filename)

        worksheet = workbook.add_worksheet()

        worksheet.write_url("A1", "http://www.perl.com/")

        # Maintain the link but overwrite string with a formula.
        worksheet.write_formula("A1", "=1+1", None, 2)

        # Reset the SST for testing.
        workbook.str_table = SharedStringTable()

        workbook.close()

        self.assertExcelEqual()
